@@ -5,13 +5,13 @@ let die = die
 type t =
   { name : string
   ; dest_dir : string
-  ; ocamlc : string
   ; log : string -> unit
   ; mutable counter : int
   ; ext_obj : string
   ; c_compiler : string
   ; stdlib_dir : string
   ; ccomp_type : string
+  ; c_libraries : string list
   ; ocamlc_config : Ocaml_config.Vars.t
   ; ocamlc_config_cmd : string
   }
@@ -19,10 +19,7 @@ type t =
 let rec rm_rf dir =
   Array.iter (Sys.readdir dir) ~f:(fun fn ->
       let fn = dir ^/ fn in
-      if Sys.is_directory fn then
-        rm_rf fn
-      else
-        Unix.unlink fn);
+      if Sys.is_directory fn then rm_rf fn else Unix.unlink fn);
   Unix.rmdir dir
 
 module Temp = struct
@@ -68,35 +65,23 @@ module Flags = struct
 end
 
 module Find_in_path = struct
-  let path_sep =
-    if Sys.win32 then
-      ';'
-    else
-      ':'
+  let path_sep = if Sys.win32 then ';' else ':'
 
   let get_path () =
     match Sys.getenv "PATH" with
     | exception Not_found -> []
     | s -> String.split s ~on:path_sep
 
-  let exe =
-    if Sys.win32 then
-      ".exe"
-    else
-      ""
+  let exe = if Sys.win32 then ".exe" else ""
 
   let prog_not_found prog = die "Program %s not found in PATH" prog
 
   let best_prog dir prog =
     let fn = dir ^/ prog ^ ".opt" ^ exe in
-    if Sys.file_exists fn then
-      Some fn
+    if Sys.file_exists fn then Some fn
     else
       let fn = dir ^/ prog ^ exe in
-      if Sys.file_exists fn then
-        Some fn
-      else
-        None
+      if Sys.file_exists fn then Some fn else None
 
   let find_ocaml_prog prog =
     match List.find_map (get_path ()) ~f:(fun dir -> best_prog dir prog) with
@@ -118,16 +103,12 @@ let gen_id t =
 
 let quote_if_needed =
   let need_quote = function
-    | ' '
-    | '\"' ->
-      true
+    | ' ' | '\"' -> true
     | _ -> false
   in
   fun s ->
-    if String.is_empty s || String.exists ~f:need_quote s then
-      Filename.quote s
-    else
-      s
+    if String.is_empty s || String.exists ~f:need_quote s then Filename.quote s
+    else s
 
 module Process = struct
   type result =
@@ -229,12 +210,10 @@ module Process = struct
 
   let run_command_capture_exn t ?dir ?env cmd =
     let { exit_code; stdout; stderr } = run_command t ?dir ?env cmd in
-    if exit_code <> 0 then
-      die "command exited with code %d: %s" exit_code cmd
+    if exit_code <> 0 then die "command exited with code %d: %s" exit_code cmd
     else if not (String.is_empty stderr) then
       die "command has non-empty stderr: %s" cmd
-    else
-      stdout
+    else stdout
 
   let run_command_ok t ?dir ?env cmd =
     (run_command t ?dir ?env cmd).exit_code = 0
@@ -310,16 +289,21 @@ let read_dot_dune_configurator_file ~build_dir =
 
 let fill_in_fields_that_depends_on_ocamlc_config t =
   let get = ocaml_config_var_exn t in
-  let c_compiler =
+  let get_flags var =
+    get var |> String.trim |> Flags.extract_blank_separated_words
+  in
+  let c_compiler, c_libraries =
     match Ocaml_config.Vars.find t.ocamlc_config "c_compiler" with
-    | Some c_comp -> c_comp ^ " " ^ get "ocamlc_cflags"
-    | None -> get "bytecomp_c_compiler"
+    | Some c_comp ->
+      (c_comp ^ " " ^ get "ocamlc_cflags", get_flags "native_c_libraries")
+    | None -> (get "bytecomp_c_compiler", get_flags "bytecomp_c_libraries")
   in
   { t with
     ext_obj = get "ext_obj"
   ; c_compiler
   ; stdlib_dir = get "standard_library"
   ; ccomp_type = get "ccomp_type"
+  ; c_libraries
   }
 
 let create_from_inside_dune ~dest_dir ~log ~build_dir ~name =
@@ -334,7 +318,6 @@ let create_from_inside_dune ~dest_dir ~log ~build_dir ~name =
   let ocamlc_config_cmd = Process.command_line ocamlc [ "-config" ] in
   fill_in_fields_that_depends_on_ocamlc_config
     { name
-    ; ocamlc
     ; log
     ; dest_dir
     ; counter = 0
@@ -344,10 +327,16 @@ let create_from_inside_dune ~dest_dir ~log ~build_dir ~name =
     ; c_compiler = ""
     ; stdlib_dir = ""
     ; ccomp_type = ""
+    ; c_libraries = []
     }
 
 let create ?dest_dir ?ocamlc ?(log = ignore) name =
-  match (ocamlc, Option.try_with (fun () -> Sys.getenv "INSIDE_DUNE")) with
+  let inside_dune =
+    match Sys.getenv "INSIDE_DUNE" with
+    | exception Not_found -> None
+    | n -> Some n
+  in
+  match (ocamlc, inside_dune) with
   | None, Some build_dir when build_dir <> "1" ->
     create_from_inside_dune ~dest_dir ~log ~build_dir ~name
   | _ ->
@@ -364,7 +353,6 @@ let create ?dest_dir ?ocamlc ?(log = ignore) name =
     let ocamlc_config_cmd = Process.command_line ocamlc [ "-config" ] in
     let t =
       { name
-      ; ocamlc
       ; log
       ; dest_dir
       ; counter = 0
@@ -372,6 +360,7 @@ let create ?dest_dir ?ocamlc ?(log = ignore) name =
       ; c_compiler = ""
       ; stdlib_dir = ""
       ; ccomp_type = ""
+      ; c_libraries = []
       ; ocamlc_config = Ocaml_config.Vars.of_list_exn []
       ; ocamlc_config_cmd
       }
@@ -388,8 +377,7 @@ let create ?dest_dir ?ocamlc ?(log = ignore) name =
     in
     fill_in_fields_that_depends_on_ocamlc_config { t with ocamlc_config }
 
-let need_to_compile_and_link_separately t =
-  (* Vague memory from writing the discover.ml script for Lwt... *)
+let is_msvc t =
   match t.ccomp_type with
   | "msvc" -> true
   | _ -> false
@@ -399,7 +387,6 @@ let compile_and_link_c_prog t ?(c_flags = []) ?(link_flags = []) code =
   Unix.mkdir dir 0o777;
   let base = dir ^/ "test" in
   let c_fname = base ^ ".c" in
-  let obj_fname = base ^ t.ext_obj in
   let exe_fname = base ^ ".exe" in
   Io.write_file c_fname code;
   logf t "compiling c program:";
@@ -407,22 +394,21 @@ let compile_and_link_c_prog t ?(c_flags = []) ?(link_flags = []) code =
   let run_ok args =
     Process.run_command_ok t ~dir (Process.command_args t.c_compiler args)
   in
-  let ok =
-    if need_to_compile_and_link_separately t then
-      run_ok (c_flags @ [ "-I"; t.stdlib_dir; "-c"; c_fname ])
-      && run_ok ("-o" :: exe_fname :: obj_fname :: link_flags)
-    else
-      run_ok
-        (List.concat
-           [ c_flags
-           ; [ "-I"; t.stdlib_dir; "-o"; exe_fname; c_fname ]
-           ; link_flags
-           ])
+  let output_flag =
+    if is_msvc t then [ "-Fe" ^ exe_fname ] else [ "-o"; exe_fname ]
   in
-  if ok then
-    Ok ()
-  else
-    Error ()
+  let ok =
+    run_ok
+      (List.concat
+         [ c_flags
+         ; [ "-I"; t.stdlib_dir ]
+         ; output_flag
+         ; [ c_fname ]
+         ; t.c_libraries
+         ; link_flags
+         ])
+  in
+  if ok then Ok () else Error ()
 
 let compile_c_prog t ?(c_flags = []) code =
   let dir = t.dest_dir ^/ sprintf "c-test-%d" (gen_id t) in
@@ -434,14 +420,20 @@ let compile_c_prog t ?(c_flags = []) code =
   logf t "compiling c program:";
   List.iter (String.split_lines code) ~f:(logf t " | %s");
   let ok =
+    let output_flag =
+      if is_msvc t then [ "-Fo" ^ obj_fname ] else [ "-o"; obj_fname ]
+    in
     Process.run_command_ok t ~dir
       (Process.command_args t.c_compiler
-         (c_flags @ [ "-I"; t.stdlib_dir; "-o"; obj_fname; "-c"; c_fname ]))
+         (List.concat
+            [ c_flags
+            ; [ "-I"; t.stdlib_dir ]
+            ; output_flag
+            ; [ "-c"; c_fname ]
+            ; t.c_libraries
+            ]))
   in
-  if ok then
-    Ok obj_fname
-  else
-    Error ()
+  if ok then Ok obj_fname else Error ()
 
 let c_test t ?c_flags ?link_flags code =
   match compile_and_link_c_prog t ?c_flags ?link_flags code with
@@ -466,10 +458,6 @@ module C_define = struct
       | Switch of bool
       | Int of int
       | String of string
-
-    let switch b = Switch b
-
-    let int i = Int i
   end
 
   let extract_program ?prelude includes vars =
@@ -532,28 +520,44 @@ const char *s%i = "BEGIN-%i-false-END";
   let extract_values obj_file vars =
     let values =
       Io.with_lexbuf_from_file obj_file ~f:(Extract_obj.extract [])
-      |> Int.Map.of_list_exn
+      |> List.fold_left ~init:Int.Map.empty ~f:(fun acc (key, v) ->
+             Int.Map.update acc key ~f:(function
+               | None -> Some [ v ]
+               | Some vs -> Some (v :: vs)))
     in
     List.mapi vars ~f:(fun i (name, t) ->
-        let raw_val =
+        let raw_vals =
           match Int.Map.find values i with
-          | None -> die "Unable to get value for %s" name
           | Some v -> v
+          | None -> die "Unable to get value for %s" name
+        in
+        let parse_val_or_exn f =
+          let f x =
+            match f x with
+            | Some s -> s
+            | None ->
+              die
+                "Unable to read variable %S of type %s. Invalid value %S in %s \
+                 found"
+                name (Type.name t) x obj_file
+          in
+          let vs =
+            List.map ~f:(fun x -> (x, f x)) raw_vals
+            |> List.sort_uniq ~cmp:(fun (_, x) (_, y) -> compare x y)
+          in
+          match vs with
+          | [] -> assert false
+          | [ (_, v) ] -> v
+          | vs ->
+            let vs = List.map ~f:fst vs in
+            die "Duplicate values for %s:\n%s" name
+              (vs |> List.map ~f:(sprintf "- %s") |> String.concat ~sep:"\n")
         in
         let value =
           match t with
-          | Type.Switch -> Bool.of_string raw_val |> Option.map ~f:Value.switch
-          | Int -> Int.of_string raw_val |> Option.map ~f:Value.int
-          | String -> Some (String raw_val)
-        in
-        let value =
-          match value with
-          | Some v -> v
-          | None ->
-            die
-              "Unable to read variable %S of type %s. Invalid value %S in %s \
-               found"
-              name (Type.name t) raw_val obj_file
+          | Type.Switch -> Value.Switch (parse_val_or_exn Bool.of_string)
+          | Int -> Value.Int (parse_val_or_exn Int.of_string)
+          | String -> String (parse_val_or_exn Option.some)
         in
         (name, value))
 
@@ -604,9 +608,9 @@ let which t prog =
   logf t "which: %s" prog;
   let x = Find_in_path.which prog in
   logf t "-> %s"
-    ( match x with
+    (match x with
     | None -> "not found"
-    | Some fn -> "found: " ^ quote_if_needed fn );
+    | Some fn -> "found: " ^ quote_if_needed fn);
   x
 
 module Pkg_config = struct
@@ -633,10 +637,7 @@ module Pkg_config = struct
       | None ->
         if
           String.exists package ~f:(function
-            | '='
-            | '>'
-            | '<' ->
-              true
+            | '=' | '>' | '<' -> true
             | _ -> false)
         then
           warn
@@ -658,9 +659,9 @@ module Pkg_config = struct
             sprintf "%s/opt/%s/lib/pkgconfig" (quote_if_needed prefix) package
           in
           Option.some_if
-            ( match Sys.is_directory p with
+            (match Sys.is_directory p with
             | s -> s
-            | exception Sys_error _ -> false )
+            | exception Sys_error _ -> false)
             p
         in
         new_pkg_config_path >>| fun new_pkg_config_path ->
@@ -688,8 +689,7 @@ module Pkg_config = struct
         | s -> String.extract_blank_separated_words s
       in
       Ok { libs = run "--libs"; cflags = run "--cflags" }
-    else
-      Error stderr
+    else Error stderr
 
   let query t ~package = Result.to_option @@ gen_query t ~package ~expr:None
 
@@ -713,12 +713,12 @@ let main ?(args = []) ~name f =
   let dest_dir = ref None in
   let args =
     Arg.align
-      ( [ ("-verbose", Arg.Set verbose, " be verbose")
-        ; ( "-dest-dir"
-          , Arg.String (fun s -> dest_dir := Some s)
-          , "DIR save temporary files to this directory" )
-        ]
-      @ args )
+      ([ ("-verbose", Arg.Set verbose, " be verbose")
+       ; ( "-dest-dir"
+         , Arg.String (fun s -> dest_dir := Some s)
+         , "DIR save temporary files to this directory" )
+       ]
+      @ args)
   in
   let anon s = raise (Arg.Bad (sprintf "don't know what to do with %s" s)) in
   let usage = sprintf "%s [OPTIONS]" (Filename.basename Sys.executable_name) in
@@ -728,11 +728,7 @@ let main ?(args = []) ~name f =
   try
     let t =
       create_from_inside_dune ~dest_dir:!dest_dir
-        ~log:
-          ( if !verbose then
-            prerr_endline
-          else
-            log )
+        ~log:(if !verbose then prerr_endline else log)
         ~build_dir ~name
     in
     f t
@@ -743,4 +739,4 @@ let main ?(args = []) ~name f =
     | Fatal_error msg ->
       eprintf "Error: %s\n%!" msg;
       exit 1
-    | _ -> Exn.raise_with_backtrace exn bt )
+    | _ -> Exn.raise_with_backtrace exn bt)
